@@ -11,35 +11,19 @@
 import asyncio
 import base64
 import concurrent
+import json
 import os
 import sys
 import tempfile
 import threading
-from functools import partial
+from functools import lru_cache, partial
+from json.decoder import JSONDecodeError
 from typing import Any, Callable, Dict, List, Optional
 
 import cognite.geospatial._client
 import numpy as np
 import pyarrow as pa
 from cognite import geospatial as geospatial
-from cognite.geospatial._client import (
-    AttributeTypeDTO,
-    CoreGeometrySpatialItemDTO,
-    CreateSpatialItemsDTO,
-    DataExtractorDTO,
-    EitherIdDTO,
-    ExternalIdDTO,
-    FeatureLayersFilterDTO,
-    GeometryProjectionDTO,
-    InternalIdDTO,
-    SpatialDataDTO,
-    SpatialDataRequestDTO,
-    SpatialDatasDTO,
-    SpatialIdsDTO,
-    SpatialItemsProjectionDTO,
-    SpatialRelationshipDTO,
-    SpatialSearchRequestDTO,
-)
 from cognite.geospatial._client.rest import ApiException
 from cognite.geospatial._spatial_filter_object import SpatialFilterObject
 from cognite.geospatial._spatial_object import SpatialObject
@@ -50,10 +34,40 @@ from tornado.concurrent import Future, future_set_exc_info, is_future
 
 from ._console import in_interactive_session
 
+from cognite.geospatial._client import (  # GeometryProjectionDTO,; SpatialItemsProjectionDTO,
+    AttributeTypeDTO,
+    CoreGeometrySpatialItemDTO,
+    CreateSpatialItemsDTO,
+    DataExtractorDTO,
+    EitherIdDTO,
+    ExternalIdDTO,
+    FeatureLayersFilterDTO,
+    InternalIdDTO,
+    IntersectionQueryDTO,
+    SpatialCoverageRequestDTO,
+    SpatialDataDTO,
+    SpatialDataRequestDTO,
+    SpatialDatasDTO,
+    SpatialIdsDTO,
+    SpatialRelationshipDTO,
+    SpatialSearchRequestDTO,
+)
+
+
 TORNADO_TIMEOUT_ERROR = 599
 TORNADO_MESSAGE = "Could not get a response from the server. The server is down or timeout happens."
 
 
+class GeospatialError(Exception):
+    """Geospatial reised for internal errors.
+    Attributes:
+        message: explanation of the error
+    """
+
+    def __init__(self, message):
+        super().__init__(message)
+
+
 def _check_id(id: int):
     if id is not None and id > 9007199254740991:
         raise ValueError("Invalid value for `id`, must be a value less than or equal to `9007199254740991`")
@@ -68,10 +82,20 @@ def _check_external_id(external_id: str):
         raise ValueError("Invalid value for `external_id`, length must be less than or equal to `255`")
 
 
-def _throw_exception(ex):
+def _throw_exception(ex: ApiException):
     # check for tornado timout exception code
     if ex.status == TORNADO_TIMEOUT_ERROR:
-        raise ApiException(status=TORNADO_TIMEOUT_ERROR, reason=TORNADO_MESSAGE)
+        raise GeospatialError(message=TORNADO_MESSAGE)
+
+    if ex.body:
+        try:
+            error_json = json.loads(ex.body)
+        except JSONDecodeError:
+            # not a json
+            raise ex
+        if "error" in error_json:
+            error = error_json["error"]
+            raise GeospatialError(message=error["message"])
     raise ex
 
 
@@ -87,16 +111,6 @@ def _check_either_external_id(id: int = None, external_id: str = None):
         raise ValueError("Either id or external_id must be provided")
 
 
-def _create_projection(projection: str = None):
-    if projection is None or projection == "2d":
-        proj = GeometryProjectionDTO._2D
-    elif projection == "3d":
-        proj = GeometryProjectionDTO._3D
-    else:
-        raise ValueError("Projection must be 2d or 3d")
-    return proj
-
-
 def _first_item(response):
     if response is None or response.items is None or len(response.items) == 0:
         return None
@@ -252,41 +266,60 @@ class CogniteGeospatialClient:
             raise TimeoutError("Operation timed out after %s seconds" % timeout)
         return future_cell[0].result()
 
-    async def get_spatial_info_async(self, id: int = None, external_id: str = None):
+    def _first_spatial_item(self, response) -> Optional[SpatialObject]:
+        if response is None or response.items is None or len(response.items) == 0:
+            return None
+        return SpatialObject(client=self, spatial_item=response.items[0])
+
+    async def get_spatial_info_async(self, id: int = None, external_id: str = None) -> Optional[SpatialObject]:
         """Retrieves spatial item information by internal ids or external ids.
         """
         spatial_by_ids = _create_spatial_ids(id, external_id)
         try:
             response = await self.api.by_ids_spatial_items(self.project, spatial_by_ids, _request_timeout=self.timeout)
-            return _first_item(response)
+            return self._first_spatial_item(response)
         except ApiException as e:
-            _throw_exception(e)
+            return _throw_exception(e)
 
-    def get_spatial_info(self, id: int = None, external_id: str = None):
+    def get_spatial_info(self, id: int = None, external_id: str = None) -> Optional[SpatialObject]:
         """Retrieves spatial item information by internal ids or external ids.
+
+        Args:
+            id (int): the id of the spatial object
+            external_id (str): the external_id reference of the spatial object
+
+        Returns:
+            FullSpatialItemDTO: A spatial object
         """
         run_func = partial(self.get_spatial_info_async, id, external_id)
         item = self._run_sync(run_func, self.timeout)
         return item
 
-    async def delete_spatial_async(self, id: int = None, external_id: str = None):
+    async def delete_spatial_async(self, id: int = None, external_id: str = None) -> Optional[SpatialObject]:
         """Delete spatial item by internal ids or external ids.
         """
         spatial_delete_ids = _create_spatial_ids(id, external_id)
         try:
             response = await self.api.delete_spatial(self.project, spatial_delete_ids, _request_timeout=self.timeout)
-            return _first_item(response)
+            return self._first_spatial_item(response)
         except ApiException as e:
-            _throw_exception(e)
+            return _throw_exception(e)
 
-    def delete_spatial(self, id: int = None, external_id: str = None):
+    def delete_spatial(self, id: int = None, external_id: str = None) -> Optional[SpatialObject]:
         """Delete spatial item by internal ids or external ids.
+
+        Args:
+            id (int): the id of the spatial object
+            external_id (str): the external_id reference of the spatial object
+
+        Returns:
+            FullSpatialItemsDTO: spatial object
         """
         run_func = partial(self.delete_spatial_async, id, external_id)
         item = self._run_sync(run_func, self.timeout)
         return item
 
-    async def get_spatial_async(self, id: int = None, external_id: str = None):
+    async def get_spatial_async(self, id: int = None, external_id: str = None) -> Optional[SpatialObject]:
         """Retrieves spatial item data by internal ids or external ids.
         """
         _check_either_external_id(id, external_id)
@@ -299,34 +332,11 @@ class CogniteGeospatialClient:
         if spatial_item is None:
             return None
 
-        geometry = None
-        attributes = spatial_item.attributes
-        if attributes is not None:
-            if "geometry" in attributes:
-                geometry = attributes["geometry"]
-            elif "coverage" in attributes:
-                geometry = attributes["coverage"]
-        spatial_object = SpatialObject(
-            client=self,
-            id=spatial_item.id,
-            external_id=spatial_item.external_id,
-            name=spatial_item.name,
-            description=spatial_item.description,
-            source=spatial_item.source,
-            crs=spatial_item.crs,
-            metadata=spatial_item.metadata,
-            layer=spatial_item.layer,
-            asset_ids=spatial_item.asset_ids,
-            geometry=geometry,
-            last_updated_time=spatial_item.last_updated_time,
-            created_time=spatial_item.created_time,
-        )
-
         try:
-            layer_filter = FeatureLayersFilterDTO(names=[spatial_item.layer])
-            response = await self.api.find_feature_layer(self.project, layer_filter, _request_timeout=self.timeout)
-            layer = _first_item(response)
+            layer = await self.get_layer_async(name=spatial_item.layer)
             if layer is not None:
+                spatial_item._set_layer_info(layer)
+
                 for layer_attribute in layer.attributes:
                     data_request = SpatialDataRequestDTO(spatial_id=item, attributes=[layer_attribute.name])
                     data = await self.api.get_spatial_items_data(
@@ -335,25 +345,25 @@ class CogniteGeospatialClient:
                     if data is not None and len(data) > 0:
                         for key, value in data.items():
                             if not layer_attribute.is_array:
-                                spatial_object.add_text(layer_attribute.name, value)
+                                spatial_item._add_text(layer_attribute.name, value)
                             elif layer_attribute.type == AttributeTypeDTO.DOUBLE:
                                 byte_buffer = base64.urlsafe_b64decode(value)
                                 vector = np.frombuffer(byte_buffer, dtype=">d")
-                                spatial_object.add_double(layer_attribute.name, vector)
+                                spatial_item._add_double(layer_attribute.name, vector)
                             elif layer_attribute.type == AttributeTypeDTO.INT:
                                 byte_buffer = base64.urlsafe_b64decode(value)
                                 vector = np.frombuffer(byte_buffer, dtype=">i")
-                                spatial_object.add_integer(layer_attribute.name, vector)
+                                spatial_item._add_integer(layer_attribute.name, vector)
                             elif layer_attribute.type == AttributeTypeDTO.BOOLEAN:
                                 byte_buffer = base64.urlsafe_b64decode(value)
                                 vector = np.frombuffer(byte_buffer, dtype=np.uint8)
                                 bit_array = np.unpackbits(vector, bitorder="little")
-                                spatial_object.add_boolean(layer_attribute.name, np.array(bit_array, dtype=bool))
+                                spatial_item._add_boolean(layer_attribute.name, np.array(bit_array, dtype=bool))
 
         except ApiException as e:
             _throw_exception(e)
 
-        return spatial_object
+        return spatial_item
 
     def __decode_attribute(self, value, type: AttributeTypeDTO):
         byte_buffer = base64.urlsafe_b64decode(value)
@@ -370,8 +380,15 @@ class CogniteGeospatialClient:
 
         return value
 
-    def get_spatial(self, id: int = None, external_id: str = None):
+    def get_spatial(self, id: int = None, external_id: str = None) -> Optional[SpatialObject]:
         """Retrieves spatial item data by internal ids or external ids.
+
+        Args:
+            id (int): the id of the spatial object
+            external_id (str): the external_id reference of the spatial object
+
+        Returns:
+            SpatialObject: spatial object's data
         """
         run_func = partial(self.get_spatial_async, id, external_id)
         result = self._run_sync(run_func, self.timeout)
@@ -389,8 +406,10 @@ class CogniteGeospatialClient:
         metadata: Dict[str, str] = None,
         source: str = None,
         external_id_prefix: str = None,
+        output_crs: str = None,
         limit: int = 10,
-    ):
+        offset: int = 0,
+    ) -> List[SpatialObject]:
         """Searches and returns the spatial items based on resource type content or coordinates.
         """
 
@@ -418,7 +437,6 @@ class CogniteGeospatialClient:
                 spatial_relationship, geometry, local_vars_configuration=self.configuration
             )
         spatial_search_request = SpatialSearchRequestDTO(
-            limit=limit,
             name=name,
             asset_ids=asset_ids,
             metadata=metadata,
@@ -427,37 +445,166 @@ class CogniteGeospatialClient:
             spatial_filter=spatial_filter,
             layer=layer,
             attributes=attributes,
+            output_crs=output_crs,
+            limit=limit,
+            offset=offset,
         )
 
         try:
             response = await self.api.search_spatial(
                 self.project, spatial_search_request_dto=spatial_search_request, _request_timeout=self.timeout
             )
-            return response.items if response is not None else None
+            if response is not None:
+                return [SpatialObject(client=self, spatial_item=item) for item in response.items]
+            return []
         except ApiException as e:
-            _throw_exception(e)
+            return _throw_exception(e)
+
+    def find_spatial(
+        self,
+        layer: str,
+        spatial_relationship: SpatialRelationship = None,
+        geometry: Geometry = None,
+        distance=None,
+        name: str = None,
+        asset_ids: List[int] = None,
+        attributes: List[str] = None,
+        metadata: Dict[str, str] = None,
+        source: str = None,
+        external_id_prefix: str = None,
+        output_crs: str = None,
+        limit: int = 10,
+        offset: int = 0,
+    ) -> List[SpatialObject]:
+        """Search the spatial objects based on metadata or spatial relationships.
+
+        Args:
+            layer (str): the layer to which objects belong
+            spatial_relationship (SpatialRelationship): the spatial relationship between looking objects and the input geometry
+            geometry (Geometry): the input geometry
+            name (str): the name of returned objects
+            asset_ids (List[int]): the list of asset ids which spatial objects must link to
+            attributes (List[str]): the list of attribute names are returned together with spatial objects
+            metadata (dict): a set of metadata pairs which spatial objects must have
+            source (str): the source of spatial objects
+            external_id_prefix (str): the prefix of external_id reference which spatial objects must have
+            output_crs (str): the crs of the output geometry attribute in case `geometry` attributes are queried
+            limit (int): number of objects to be returned
+            offset (int): the starting offset of objects in the results
+        Returns:
+            List[SpatialObject]: spatial object's data
+
+        Example:
+                >>> # search within the geometry of another spatial item
+                >>> client.find_spatial(layer="polygon", spatial_relationship=SpatialRelationship.within,
+                ...        geometry=Geometry(external_id="external_id"),
+                ...    )
+                >>> # search within a specific area defined by user
+                >>> items = client.find_spatial(
+                ...     layer="polygon",
+                ...     spatial_relationship=SpatialRelationship.within,
+                ...     geometry=Geometry(
+                ...         wkt="POLYGON ((289999 5999999, 290006 5999999, 290006 6000006, 289999 6000006, 289999 5999999))",
+                ...         crs="epsg:23031",
+                ...         ),
+                ...     )
 
-    async def get_coverage_async(self, id: int = None, external_id: str = None, projection: str = None):
-        """Retrieves spatial item information by internal ids or external ids.
         """
-        spatial_by_ids = _create_spatial_ids(id, external_id)
-        proj = _create_projection(projection)
-        spatialite_projection = SpatialItemsProjectionDTO(ids=spatial_by_ids, projection=proj)
+        run_func = partial(
+            self.find_spatial_async,
+            layer,
+            spatial_relationship,
+            geometry,
+            distance,
+            name,
+            asset_ids,
+            attributes,
+            metadata,
+            source,
+            external_id_prefix,
+            output_crs,
+            limit,
+            offset,
+        )
+        result = self._run_sync(run_func, self.timeout)
+        return result
+
+    async def get_coverage_async(
+        self, output_crs: str, id: int = None, external_id: str = None, dimensional_space: str = "2d"
+    ):
+        """Retrieves the coverage of the spatial object by internal ids or external ids.
+        """
+        item_id = EitherIdDTO(id, external_id, local_vars_configuration=self.configuration)
+        spatial_coverage_request = SpatialCoverageRequestDTO(
+            items=[item_id],
+            dimensional_space=dimensional_space,
+            output_crs=output_crs,
+            local_vars_configuration=self.configuration,
+        )
         try:
             response = await self.api.get_spatial_coverage(
-                self.project, spatialite_projection, _request_timeout=self.timeout
+                self.project, spatial_coverage_request, _request_timeout=self.timeout
             )
             return _first_item(response)
         except ApiException as e:
             _throw_exception(e)
 
-    def get_coverage(self, id: int = None, external_id: str = None, projection: str = None):
-        """Retrieves spatial item information by internal ids or external ids.
+    def get_coverage(self, output_crs: str, id: int = None, external_id: str = None, dimensional_space: str = "2d"):
+        """Retrieve the coverage of the spatial object by internal ids or external ids.
+
+        Args:
+            id (int): the id of the spatial object
+            external_id (str): the external_id reference of the spatial object
+            dimensional_space (str): The spatial dimension of the coverage. Valid values are "2d", "3d"
+            output_crs (str): the crs of the output coverage
+
+        Returns:
+            SpatialItemsCoveragesDTO: spatial object's data
         """
-        run_func = partial(self.get_coverage_async, id, external_id, projection)
+        run_func = partial(self.get_coverage_async, output_crs, id, external_id, dimensional_space)
         item = self._run_sync(run_func, self.timeout)
         return item
 
+    async def list_layers_async(self):
+        layer_filter = FeatureLayersFilterDTO(names=[])
+        try:
+            response = await self.api.find_feature_layer(self.project, layer_filter, _request_timeout=self.timeout)
+            return response.items
+        except ApiException as e:
+            _throw_exception(e)
+
+    def list_layers(self):
+        """List all available layers.
+
+        Returns:
+            list[FeatureLayerDTO]: list of layers
+        """
+        run_func = partial(self.list_layers_async)
+        items = self._run_sync(run_func, self.timeout)
+        return items
+
+    async def get_layer_async(self, name: str):
+        layer_filter = FeatureLayersFilterDTO(names=[name])
+        try:
+            response = await self.api.find_feature_layer(self.project, layer_filter, _request_timeout=self.timeout)
+            return _first_item(response)
+        except ApiException as e:
+            _throw_exception(e)
+
+    @lru_cache(maxsize=128)
+    def get_layer(self, name: str):
+        """Get layer by name.
+
+        Args:
+            name (str): name of the layer
+
+        Returns:
+            FeatureLayerDTO: layer information
+        """
+        run_func = partial(self.get_layer_async, name)
+        layer = self._run_sync(run_func, self.timeout)
+        return layer
+
     def get_attributes(
         self,
         layer_name: str,
@@ -469,9 +616,7 @@ class CogniteGeospatialClient:
         async def get_attributes(
             layer_name: str, item_id: EitherIdDTO, attributes: List[str], extractors: List[DataExtractor] = None
         ):
-            layer_filter = FeatureLayersFilterDTO(names=[layer_name])
-            response = await self.api.find_feature_layer(self.project, layer_filter, _request_timeout=self.timeout)
-            layer = _first_item(response)
+            layer = await self.get_layer_async(name=layer_name)
             if layer is None:
                 raise ValueError("layer does not exist")
 
@@ -493,38 +638,88 @@ class CogniteGeospatialClient:
         run_func = partial(get_attributes, layer_name, item_id, attributes, extractors)
         return self._run_sync(run_func, self.timeout)
 
-    def find_spatial(
-        self,
-        layer: str,
-        spatial_relationship: SpatialRelationship = None,
-        geometry: Geometry = None,
-        distance=None,
-        name: str = None,
-        asset_ids: List[int] = None,
-        attributes: List[str] = None,
-        metadata: Dict[str, str] = None,
-        source: str = None,
-        external_id_prefix: str = None,
-        limit: int = 10,
-    ):
-        """Searches and returns the spatial items based on resource type content or coordinates.
+    # def has_spatial(
+    #     self,
+    #     layer: str,
+    #     spatial_relationship: SpatialRelationship = None,
+    #     geometry: Geometry = None,
+    #     distance=None,
+    #     name: str = None,
+    #     asset_ids: List[int] = None,
+    #     metadata: Dict[str, str] = None,
+    #     source: str = None,
+    #     external_id_prefix: str = None,
+    # ):
+    #     """Test if spatial objects exist based on metadata or spatial relationships.
+
+    #     Args:
+    #         layer (str): the layer to which objects belong
+    #         spatial_relationship (SpatialRelationship): the spatial relationship between looking objects and the input geometry
+    #         geometry (Geometry): the input geometry
+    #         name (str): the name of returned objects
+    #         asset_ids (List[int]): the list of asset ids which spatial objects must link to
+    #         metadata (dict): a set of metadata pairs which spatial objects must have
+    #         source (str): the source of spatial objects
+    #         external_id_prefix (str): the prefix of external_id reference which spatial objects must have
+    #     Returns:
+    #         bool: True - if data exists, otherwise False
+
+    #     Example:
+    #             >>> # has spatial items within the geometry of another spatial item
+    #             >>> client.has_spatial(layer="polygon", spatial_relationship=SpatialRelationship.within,
+    #             ...        geometry=Geometry(external_id="external_id"),
+    #             ...    )
+    #             >>> # has spatial items within a specific area defined by user
+    #             >>> has_polygons = client.has_spatial(
+    #             ...     layer="polygon",
+    #             ...     spatial_relationship=SpatialRelationship.within,
+    #             ...     cognite=Geometry(
+    #             ...         wkt="POLYGON ((289999 5999999, 290006 5999999, 290006 6000006, 289999 6000006, 289999 5999999))",
+    #             ...         crs="epsg:23031",
+    #             ...         ),
+    #             ...     )
+
+    #     """
+    #     items = self.find_spatial(
+    #         layer=layer,
+    #         spatial_relationship=spatial_relationship,
+    #         geometry=geometry,
+    #         distance=distance,
+    #         name=name,
+    #         asset_ids=asset_ids,
+    #         attributes=None,
+    #         metadata=metadata,
+    #         source=source,
+    #         external_id_prefix=external_id_prefix,
+    #         limit=1,
+    #     )
+    #     return len(items) == 1
+
+    def get_layer_items(self, layer: str, attributes: List[str] = None, output_crs: str = None):
+        """Get all spatial items belong to layer.
+        WARNING: This method will load all spatial items into memory and has a risk to get memory overflow.
+
+        Args:
+            layer (str): the layer to which objects belong
+            attributes (List[str]): the list of attribute names are returned together with spatial objects
+            output_crs (str): the crs of the output geometry attribute in case `geometry` attributes are queried
+
+        Returns:
+            DenseSpatialItemDTO: spatial object's data
         """
-        run_func = partial(
-            self.find_spatial_async,
-            layer,
-            spatial_relationship,
-            geometry,
-            distance,
-            name,
-            asset_ids,
-            attributes,
-            metadata,
-            source,
-            external_id_prefix,
-            limit,
-        )
-        result = self._run_sync(run_func, self.timeout)
-        return result
+        items = []
+        offset = 0
+        limit = 100
+        while True:
+            bucket = self.find_spatial(
+                layer=layer, attributes=attributes, limit=limit, offset=offset, output_crs=output_crs
+            )
+            items.extend(bucket)
+            offset += limit
+            if limit != len(bucket):
+                break
+
+        return items
 
     def create_geometry(
         self,
@@ -538,6 +733,22 @@ class CogniteGeospatialClient:
         geometry: str = None,
         asset_ids: List[int] = None,
     ):
+        """ Create a geometry with its metadata
+
+        Args:
+            name (str): geometry name
+            external_id (str): external_id reference
+            description (str): the description
+            metadata (dict): metadata of geometry
+            layer (str): the layer to which the geometry belongs
+            source (str): the source of geometry
+            crs (str): the crs on which the geometry is defined
+            geometry (str): wkt form of the geometry
+            asset_ids (List[int]): a list of associated asset ids
+
+        Returns:
+            FullSpatialItemDTO: A spatial object
+        """
         attributes = {"geometry": geometry}
         run_func = partial(
             self.create_spatial_async,
@@ -665,56 +876,66 @@ class CogniteGeospatialClient:
         try:
             response = await self.api.file_save(
                 self.project,
+                file_type="parquet",
+                file=file,
                 id=id,
                 external_id=external_id,
                 attributes=attributes,
-                file_type="parquet",
-                file=file,
                 _request_timeout=self.file_timeout,
             )
             return response
         except ApiException as e:
             _throw_exception(e)
 
-    async def shape_file_save_async(self, file, layer: str, attributes=None):
+    async def shape_file_save_async(self, file, layer: str, create_layer: bool = False, attributes=None):
         """Shapefile save as spatial items.
         """
         try:
             response = await self.api.file_save(
                 self.project,
-                layer=layer,
-                attributes=attributes,
                 file_type="shp",
                 file=file,
+                layer=layer,
+                create_layer=create_layer,
+                attributes=attributes,
                 _request_timeout=self.file_timeout,
             )
             return response
         except ApiException as e:
             _throw_exception(e)
 
-    def shape_file_save(self, file, layer: str, attributes=None):
-        """Shapefile save as spatial items.
+    def shape_file_save(self, file, layer: str, create_layer: bool = False, attributes=None):
+        """Extract spatial items from shapefile and save it into geospatial.
+
+        Args:
+            file (str): file location on local drive
+            layer (str): the layer name to save
+            create_layer (bool): create layer if not exists
+            attributes Optional(List[str]): attributes to save (if None save all attributes)
         """
-        run_func = partial(self.shape_file_save_async, file, layer, attributes)
+        run_func = partial(self.shape_file_save_async, file, layer, create_layer, attributes)
         result = self._run_sync(run_func, self.file_timeout)
         return result
 
-    async def calculate_spatial_coverage_async(self, id: int = None, external_id: str = None):
-        """Calculate spatial item coverage by internal ids or external ids.
+    async def interpret_async(self, id: int = None, external_id: str = None):
+        """Interpret attributes from spatial item by internal ids or external ids.
         """
         spatial_by_ids = _create_spatial_ids(id, external_id)
         try:
-            response = await self.api.calculate_spatial_coverage(
-                self.project, spatial_by_ids, _request_timeout=self.timeout
-            )
+            response = await self.api.interpret(self.project, spatial_by_ids, _request_timeout=self.timeout)
             return response
         except ApiException as e:
             _throw_exception(e)
 
-    def calculate_spatial_coverage(self, id: int = None, external_id: str = None):
-        """Calculate spatial item coverage by internal ids or external ids.
+    def interpret(self, id: int = None, external_id: str = None):
+        """Interpret attributes from spatial item by internal ids or external ids.
+
+        Args:
+            id (int): the id of the spatial object
+            external_id (str): the external_id reference of the spatial object
+
         """
-        run_func = partial(self.calculate_spatial_coverage_async, id, external_id)
+        run_func = partial(self.interpret_async, id, external_id)
         result = self._run_sync(run_func, self.timeout)
         return result
 
@@ -738,9 +959,7 @@ class CogniteGeospatialClient:
 
         name_attribute_map = {}
         if attributes is not None:
-            layer_filter = FeatureLayersFilterDTO(names=[layer])
-            response = await self.api.find_feature_layer(self.project, layer_filter)
-            layer_data = _first_item(response)
+            layer_data = await self.get_layer_async(name=layer)
             if layer_data is None:
                 raise ValueError(f"layer {layer} does not exist")
 
@@ -759,14 +978,14 @@ class CogniteGeospatialClient:
 
         if attributes is not None:
             for name, value in attributes.items():
-                if not _is_primitive(value):
+                if name in name_attribute_map and not _is_primitive(value):
                     attribute = name_attribute_map.get(name)
 
                     await self.store_spatial_item_data_async(item.id, name, value, attribute.type)
 
                     item.attributes[name] = value
 
-        await self.calculate_spatial_coverage_async(id=item.id)
+        await self.interpret_async(id=item.id)
         return item
 
     def save_spatial(
@@ -774,15 +993,78 @@ class CogniteGeospatialClient:
         name: str = None,
         external_id: str = None,
         description: str = None,
-        metadata=None,
+        metadata: Dict[str, str] = None,
         layer: str = None,
         source: str = None,
         crs: str = None,
-        attributes=None,
+        attributes: dict = None,
         asset_ids: List[int] = None,
     ):
+        """ Create spatial object with additional attributes
+
+        Args:
+            name (str): object name
+            external_id (str): external_id reference
+            description (str): the description
+            metadata (dict): metadata of the object
+            layer (str): the layer to which the object belongs
+            source (str): the source of the object
+            crs (str): the crs where geometry attributes are defined
+            attributes (dict): a map of attribute's name and value
+            asset_ids (List[int]): a list of associated asset ids
+        Returns:
+            FullSpatialItemDTO: A spatial object
+        """
         run_func = partial(
             self.save_spatial_async, name, external_id, description, metadata, layer, source, crs, attributes, asset_ids
         )
         result = self._run_sync(run_func, self.timeout)
         return result
+
+    def get_intersections(self, geometry: Geometry, geometries: List[Geometry], output_crs):
+        """ Find intersection between an array of geometries and another geometry
+
+        Args:
+            geometry (Geometry): a single geometry
+            geometries (List[Geometry]): an array of geometries
+            output_crs (str): the crs of intersection geometries
+        Returns:
+            List[str]: a list of intersections, each in the form of wkt
+
+        Example:
+                >>> intersections = client.get_intersections(
+                ...         output_crs="epsg:4326",
+                ...         geometry=Geometry(external_id="external_id_1"),
+                ...         geometries=[
+                ...             Geometry(external_id="external_id_2"),
+                ...             Geometry(external_id="external_id_3"),
+                ...             Geometry(external_id="external_id_4"),
+                ...         ],
+                ...     )
+
+        """
+
+        async def get_intersections_async(geometry: Geometry, geometries: List[Geometry], output_crs):
+            geometry_dto = EitherIdDTO(
+                id=geometry.id, external_id=geometry.external_id, local_vars_configuration=self.configuration
+            )
+            geometries_dto = [
+                EitherIdDTO(id=geom.id, external_id=geom.external_id, local_vars_configuration=self.configuration)
+                for geom in geometries
+            ]
+            intersection_query_dto = IntersectionQueryDTO(
+                geometry=geometry_dto, geometries=geometries_dto, output_crs=output_crs
+            )
+            try:
+                response = await self.api.find_intersection(
+                    self.project, intersection_query_dto, _request_timeout=self.timeout
+                )
+                if response is None:
+                    return None
+                return [short_wkt.wkt for short_wkt in response.geometries]
+            except ApiException as e:
+                _throw_exception(e)
+
+        run_func = partial(get_intersections_async, geometry, geometries, output_crs)
+        result = self._run_sync(run_func, self.timeout)
+        return result
