@@ -17,10 +17,12 @@ package_data = \
 install_requires = \
 ['cognite-geospatial-sdk>=0.1.8,<0.2.0',
  'deprecation>=2.1.0,<3.0.0',
- 'grpcio-tools>=1.24,<2.0',
+ 'grpcio-tools>=1.31,<2.0',
  'numpy>=1.17,<2.0',
+ 'pip>=20.2.2,<21.0.0',
  'protobuf>=3.12,<4.0',
  'python-dateutil>=2.8,<3.0',
+ 'setuptools>=49.0,<50.0',
  'six>=1.14,<2.0',
  'urllib3>=1.24,<2.0']
 
@@ -29,7 +31,7 @@ extras_require = \
 
 setup_kwargs = {
     'name': 'cognite-seismic-sdk',
-    'version': '0.1.58',
+    'version': '0.1.59',
     'description': '',
     'long_description': None,
     'author': 'cognite',
